#include <stdlib.h>
#include <gba.h>
#include <screen_mode.h>
#include <interrupt.h>
#include "it_handlers.h"
#include "shapes.h"
#include "game_core.h"
#include "key_control.h"
#include "ihm.h"
#include "gfx_effects.h"
#include "shapes_gfx.h"
#include "cells_gfx.h"
#include "small_digits_gfx.h"
#include "large_digits_gfx.h"
#include "bg1_tiles_layout.h"
#include "shapes_sprites_gfx.h"
#include "shapes_configurations.h"
#include "pause.h"
#include "levels_data.h"
#include "wav_dontfit.h"
#include "wav_fit.h"
#include "wav_no.h"


static struct game global_game ;





struct game* start_game (struct aim_data *aim_data)
{
  unsigned short x, y ;
  u16 raw_data ;
  u8 deflated_data ;

  /* On vide le tableau. */
  for (x = 0; x < BOARD_WIDTH; x++) {
    for (y = 0; y < BOARD_HEIGHT; y++)
      global_game.board[x][y] = Shape_None ;
  }

  /* On prpare le dessin  obtenir. */
  global_game.nb_remaining_cells = 0 ;
  for (y = 0; y < BOARD_HEIGHT; y++) {
    raw_data = aim_data->bits[y] ;
    for (x = 0; x < BOARD_WIDTH; x++) {
      deflated_data = ((raw_data & (1 << ((BOARD_WIDTH - 1) - x))) ? 1 : 0) ;
      global_game.aim[x][y] = deflated_data ;
      global_game.nb_remaining_cells += deflated_data ;
    }
  }

  /* Mise  jour du nombre de pices disponibles. */
  for (x = 0; x < SHAPES_NUMBER ; x++)
    global_game.remaining_pieces[x] = aim_data->pieces[x] ;

  /* Au dbut le joueur ne tient aucune pice. */
  global_game.state = Free_hand ;
  global_game.player_location.player_area = Area_Select_Shape_Square_4x4 ;
  /* On initialise les coordonnes de retour en zone de jeu par dfaut */
  /* au milieu de cette zone. */
  global_game.player_location.x_as_index = (BOARD_WIDTH) / 2 ;
  global_game.player_location.y_as_index = (BOARD_HEIGHT) / 2 ;

  /* Remise  0 du compteur de secondes. */
  elapsed_seconds = 0 ;
  /* Mise en route du timer 2. */

  return (&global_game) ;
}




static void clamp_player_location (struct game *game)
{
  struct location *location ;
  enum orientation tmp_orient ;
  enum shape tmp_shape ;
  struct chain_code *tmp_ccode = NULL ; ;
  short x, y, x_margin, y_margin ;
  short i ;

  location = &game->player_location ;
  if (location->player_area != Area_board) return ;
  if (game->state == Free_hand) {
    /* Dans ce cas le joueur ne tient pas de pice et donc on doit */
    /* juste le restreindre aux dimensions de l'espace de jeu.     */
    /* Bound des Xs. */
    if (location->x_as_index >= BOARD_WIDTH)
      location->x_as_index = BOARD_WIDTH - 1 ;
    else
      if (location->x_as_index < 0)
	location->x_as_index = 0 ;
    /* Bound des Ys. */
    if (location->y_as_index >= BOARD_HEIGHT)
      location->y_as_index = BOARD_HEIGHT - 1 ;
    else
      if (location->y_as_index < 0)
	location->y_as_index = 0 ;
    return ;
  }

  tmp_orient = GET_ORIENT (location->handled_piece) ;
  /* On dcrmente le code de la forme car 0 est utilis pour   */
  /* dire "pas de pice" or tous les tableaux de configurations */
  /* sont indexs sur le code des pices, mais  partir de 0.   */
  tmp_shape = (GET_SHAPE (location->handled_piece)) - 1;
  tmp_ccode =
    (struct chain_code*) rotated_shapes_ccodes[tmp_shape][tmp_orient] ;

  x = location->x_as_index ;
  y = location->y_as_index ;

  for (i = 0; i < tmp_ccode->length; i++) {
    /* Correction du x  gauche. */
    if (x < 0) {
      location->x_as_index += (-x)  ;
      x = 0 ;
    }
    /* Correction du x  droite. */
    x_margin = x - (BOARD_WIDTH - 1) ;
    if (x_margin > 0) {
      location->x_as_index -= x_margin ;
      x -= x_margin ;
    }
    /* Correction du y en haut. */
    if (y < 0) {
      location->y_as_index += (-y) ;
      y = 0 ;
    }
    /* Correction du y en bas. */
    y_margin = y - (BOARD_HEIGHT - 1) ;
    if (y_margin > 0) {
      location->y_as_index -= y_margin ;
      y -= y_margin ;
    }

    x = x + tmp_ccode->codes[i].x_delta ;
    y = y + tmp_ccode->codes[i].y_delta ;
  }
}




static short try_lay_down_piece (struct game *game)
{
  enum orientation current_orient ;
  enum shape current_shape ;
  struct chain_code *current_ccode = NULL ;
  short i, x, y ;
  u16 *output_map = SCREEN_BASE_BLOCK_9_ADDR ;
  const u8 *tmp_map ;

  current_orient = GET_ORIENT(game->player_location.handled_piece) ;
  current_shape = (GET_SHAPE(game->player_location.handled_piece)) ;
  /* On va rcuprer le chain code de la pice actuellement tenue. */
  /* On utilise le code de la forme -1 car 0 est utilis pour   */
  /* dire "pas de pice" or tous les tableaux de configurations */
  /* sont indexs sur le code des pices, mais  partir de 0.   */
  current_ccode =
    (struct chain_code*)
    rotated_shapes_ccodes[current_shape - 1][current_orient] ;

  x = game->player_location.x_as_index ;
  y = game->player_location.y_as_index ;
  for (i = 0; i < current_ccode->length; i++) {
    if (((GET_SHAPE (game->board[x][y])) != Shape_None) ||
	(game->aim[x][y] == 0))
      return 0 ;             /* Espace non autoris. */
    else {
      x = x + current_ccode->codes[i].x_delta ;
      y = y + current_ccode->codes[i].y_delta ;
    }
  }

  /* On a russi  loger la pice, on met donc  jour l'aire de jeu. */
  tmp_map = rotated_shapes_maps[current_shape - 1][current_orient] ;
  x = game->player_location.x_as_index ;
  y = game->player_location.y_as_index ;
  for (i = 0; i < current_ccode->length; i++) {
    SET_PIECE(game->board[x][y], i, current_orient, current_shape) ;
    /* En mme temps on met  jour l'affichage du layer 1. */
    output_map [((y + Area_board_TOP_Y) * (256/8)) + (Area_board_TOP_X + x)] =
      (u16) ((SHAPES_START_INDEX + tmp_map[i]) | (SHAPES_PALETTE_SLOT << 12)) ;
    /* Morceau suivant de la pice. */
    x = x + current_ccode->codes[i].x_delta ;
    y = y + current_ccode->codes[i].y_delta ;
  }

  /* On dcrmente le nombre de cases restant  remplir. */
  game->nb_remaining_cells -= current_ccode->length ;
  return (1) ;
}




static short try_select_piece_on_board (struct game *game)
{
  enum orientation current_orient ;
  enum shape current_shape ;
  struct chain_code *current_ccode = NULL ;
  u16 *output_map = SCREEN_BASE_BLOCK_9_ADDR ;
  short i, x, y, part ;
  
  x = game->player_location.x_as_index ;
  y = game->player_location.y_as_index ;
  current_shape = GET_SHAPE (game->board[x][y]) ;
  if (current_shape == Shape_None) return (0) ;

  current_orient = GET_ORIENT(game->board[x][y]) ;
  /* Mise  jour du joueur. */
  SET_PIECE (game->player_location.handled_piece,
	     0, current_orient, current_shape) ;
  game->state = Handling_piece ;

  /* On dcrmente le code de la forme car 0 est utilis pour   */
  /* dire "pas de pice" or tous les tableaux de configurations */
  /* sont indexs sur le code des pices, mais  partir de 0.   */
  current_shape-- ;

  /* On va rcuprer le chain code de la pice du damier. */
  current_ccode =
    (struct chain_code*) rotated_shapes_ccodes[current_shape][current_orient] ;
  part = GET_PART(game->board[x][y]) ;
  for (i = 0; i < current_ccode->length; i++) {
    UNSET_PIECE (game->board[x][y]) ;
    /* En mme temps on met  jour l'affichage du layer 1. */
    output_map [((y + Area_board_TOP_Y) * (256/8)) + (Area_board_TOP_X + x)] =
      (u16) (CELL_EMPTY_INDEX | (SHAPES_PALETTE_SLOT << 12)) ;
    x = x + current_ccode->codes[part].x_delta ;
    y = y + current_ccode->codes[part].y_delta ;
    part = (part + 1) % current_ccode->length ;
  }

  game->nb_remaining_cells += current_ccode->length ;
  clamp_player_location (game) ;
  return (1) ;
}





static short start_button_pressed (struct game *game)
{
  short ret_val ;

  hide_player_sprites () ;
  ret_val = invoke_pause_menu () ;
  /* Il faut restaurer la palette normale du jeu. */
  restore_game_palette () ;
  refresh_player_cursor (game) ;
  return (ret_val) ;
}





static short l_or_r_button_pressed (u8 is_l, struct game *game)
{
  short ret_val = 0 ;
  enum orientation tmp_orient ;

  switch (game->player_location.player_area) {
  case Area_board:
    switch (game->state) {
    case Free_hand:
      /* Aucun effet si pas de pice tenue. */
      break ;                  /* Fin du cas Free_hand. */

    case Handling_piece:
      tmp_orient = GET_ORIENT(game->player_location.handled_piece) ;
      if (is_l) {
	/* Rotation en sens inverse des aiguille d'une montre. */
	if (tmp_orient == Orient_Up) tmp_orient = Orient_Left ;
	else tmp_orient-- ;
      }
      else {
	/* Rotation en sens des aiguille d'une montre. */
	tmp_orient = (tmp_orient + 1) % 4;
      }
      SET_ORIENT(game->player_location.handled_piece, tmp_orient) ;
      ret_val = 1 ;
      break ;                  /* Fin du cas Handling_piece. */
    }                          /* Fin de switch (game->state). */
    clamp_player_location (game) ;
    break ;                    /* Fin du cas Area_board. */

  case Area_Select_Shape_Square_1x1:
  case Area_Select_Shape_Square_4x4:
  case Area_Select_Shape_Long:
  case Area_Select_Shape_Junction:
  case Area_Select_Shape_Stair_1:
  case Area_Select_Shape_Stair_2:
  case Area_Select_Shape_Corner:
    /* Aucun effet en phase de slection de pice. */
    break ;                    /* Fin des cas de Area_Select_Shape_XX. */
  }                     /* Fin de witch (game->player_location.player_area). */

  return (ret_val) ;
}




/* Retourne 0 si le dplacement a chou. */
static short move_cursor_down (struct game *game)
{
  short ret_val = 0 ;
  short old_x, old_y ;

  switch (game->player_location.player_area) {
  case  Area_board:
    old_x = game->player_location.x_as_index ;
    old_y = game->player_location.y_as_index ;
    game->player_location.y_as_index++ ;
    clamp_player_location (game) ;
    if ((old_x != game->player_location.x_as_index) ||
	(old_y != game->player_location.y_as_index))
      ret_val = 1 ;
    break ;

  case Area_Select_Shape_Junction:
    game->player_location.player_area = Area_Select_Shape_Square_1x1 ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Square_1x1:
    game->player_location.player_area = Area_Select_Shape_Stair_1 ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Stair_1:
    game->player_location.player_area = Area_Select_Shape_Square_4x4 ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Square_4x4:
    game->player_location.player_area = Area_Select_Shape_Stair_2 ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Stair_2:
    game->player_location.player_area =  Area_Select_Shape_Long ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Long:
    game->player_location.player_area = Area_Select_Shape_Corner ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Corner:
    game->player_location.player_area = Area_Select_Shape_Junction ;
    ret_val = 1 ;
    break;
   }

  return (ret_val) ;
}





/* Retourne 0 si le dplacement a chou. */
static short move_cursor_up (struct game *game)
{
  short ret_val = 0 ;
  short old_x, old_y ;

  switch (game->player_location.player_area) {
  case  Area_board:
    old_x = game->player_location.x_as_index ;
    old_y = game->player_location.y_as_index ;
    game->player_location.y_as_index-- ;
    clamp_player_location (game) ;
    if ((old_x != game->player_location.x_as_index) ||
	(old_y != game->player_location.y_as_index))
      ret_val = 1 ;
    break ;

  case Area_Select_Shape_Junction:
    game->player_location.player_area = Area_Select_Shape_Corner ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Square_1x1:
    game->player_location.player_area = Area_Select_Shape_Junction ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Stair_1:
    game->player_location.player_area = Area_Select_Shape_Square_1x1 ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Square_4x4:
    game->player_location.player_area = Area_Select_Shape_Stair_1 ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Stair_2:
    game->player_location.player_area = Area_Select_Shape_Square_4x4 ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Long:
    game->player_location.player_area = Area_Select_Shape_Stair_2 ;
    ret_val = 1 ;
    break;

  case Area_Select_Shape_Corner:
    game->player_location.player_area = Area_Select_Shape_Long ;
    ret_val = 1 ;
    break;
  }

  return (ret_val) ;
}





/* Retourne 0 si le dplacement a chou. */
static short move_cursor_left (struct game *game)
{
  short ret_val = 0 ;
  short old_x, old_y ;

  switch (game->player_location.player_area) {
  case  Area_board:
    old_x = game->player_location.x_as_index ;
    old_y = game->player_location.y_as_index ;
    game->player_location.x_as_index-- ;
    clamp_player_location (game) ;
    if ((old_x != game->player_location.x_as_index) ||
	(old_y != game->player_location.y_as_index))
      ret_val = 1 ;
    break ;

  default:
    /* Si on n'est pas en zone de jeu, a ne fait rien. */
    ret_val = 0 ;
    break ;
  }

  return (ret_val) ;
}





/* Retourne 0 si le dplacement a chou. */
static short move_cursor_right (struct game *game)
{
  short ret_val = 0 ;
  short old_x, old_y ;

  switch (game->player_location.player_area) {
  case  Area_board:
    old_x = game->player_location.x_as_index ;
    old_y = game->player_location.y_as_index ;
    game->player_location.x_as_index++ ;
    clamp_player_location (game) ;
    if ((old_x != game->player_location.x_as_index) ||
	(old_y != game->player_location.y_as_index))
      ret_val = 1 ;
    break ;

  default:
    /* Si on n'est pas en zone de jeu, a ne fait rien. */
    ret_val = 0 ;
    break ;
  }

  return (ret_val) ;
}





static short b_button_pressed (struct game *game)
{
  short ret_val = 0 ;
  enum shape tmp_shape ;

  switch (game->player_location.player_area) {
  case Area_board:
    /* Le joueur est en zone de jeu. */
    switch (game->state) {
    case Free_hand:
      /* Il veut alors aller slectionner une pice dans le pool. */
      game->player_location.player_area = Area_Select_Shape_Square_4x4 ;
      ret_val = 1 ;
      break ;

    case Handling_piece:
      /* Il tenait une pice qu'il veut remettre dans le pool.              */
      /* Par contre, il reste au dessus de l'aire de jeu en mode Free_hand. */
      game->state = Free_hand ;
      tmp_shape = GET_SHAPE(game->player_location.handled_piece) ;
      game->remaining_pieces[tmp_shape - 1]++ ;
      refresh_shape_count (game, tmp_shape) ;
      ret_val = 1 ;
      break ;
    }                           /* Fin de switch (game->state). */
    break ;                     /* Fin du cas Area_board. */

  case Area_Select_Shape_Square_1x1:
  case Area_Select_Shape_Square_4x4:
  case Area_Select_Shape_Long:
  case Area_Select_Shape_Junction:
  case Area_Select_Shape_Stair_1:
  case Area_Select_Shape_Stair_2:
  case Area_Select_Shape_Corner:
    /* Le joueur est en zone de slection de pice, donc il  */
    /* veut retourner en aire de jeu mais en mode Free_hand. */
    game->state = Free_hand ;
    game->player_location.player_area = Area_board ;
    ret_val = 1 ;
    break ;              /* Fin des cas de Area_Select_Shape_XX. */
  }                 /* Fin de switch (game->player_location.player_area) */

  return (ret_val) ;
}





static short a_button_pressed (struct game *game)
{
  short ret_val = 0 ;

  switch (game->player_location.player_area) {
  case Area_board:
    /* Le joueur est en zone de jeu. */
    switch (game->state) {
    case Free_hand:
      /* Il veut slectionner une pice du damier. */
      ret_val = try_select_piece_on_board (game) ;
      if (ret_val) ret_val = 2 ;  /* On fait clignoter la pice au retour. */
      break ;                   /* Fin du cas Free_hand. */

    case Handling_piece:
      /* Il tenait une pice qu'il veut poser. */
      ret_val = try_lay_down_piece (game) ;
      if (ret_val) {
	/* Le joueur a russi  poser sa pice. Il n'en tient donc plus. */
	play_sound (g_wav_fit) ;
	game->state = Free_hand ;
	game->player_location.player_area = Area_Select_Shape_Square_4x4 ;
      }
      else {
	/* Le joueur n'a pas russi  poser sa pice on la secoue. */
	play_sound (g_wav_dontfit) ;
	shake_player_shape (game) ;
      }
      break ;                   /* Fin du cas Handling_piece. */
    }                           /* Fin de switch (game->state). */
    break ;                     /* Fin du cas Area_board. */

  /* Dans les cas ci-dessous, le joueur est en zone de slection de pice. */
  case Area_Select_Shape_Square_1x1:
    if (game->remaining_pieces[Shape_Square_1x1 - 1] > 0) {
      /* Il reste une pice de moins. */
      game->remaining_pieces[Shape_Square_1x1 - 1]-- ;
      /* On renvoie le joueur dans la zone de jeu. */
      game->player_location.player_area = Area_board ;
      game->player_location.hot_spot = Default_Shape_Square_1x1_Hot_Spot ;
      /* Et maintenant, le joueur tient une pice. */
      game->state = Handling_piece ;
      SET_PIECE (game->player_location.handled_piece,
		 0, Orient_Up, Shape_Square_1x1) ;
      clamp_player_location (game) ;
      refresh_shape_count (game, Shape_Square_1x1) ;
      ret_val = 2 ;         /* On fait clignoter la pice au retour. */
    }
    else {
      play_sound (g_wav_no) ;
      shake_and_blink_screen () ;
    }
    break ;                  /* Fin du cas Area_Select_Shape_Square_1x1. */
    
  case Area_Select_Shape_Square_4x4:
    if (game->remaining_pieces[Shape_Square_4x4 - 1] > 0) {
      game->remaining_pieces[Shape_Square_4x4 - 1]-- ;
      game->player_location.player_area = Area_board ;
      game->player_location.hot_spot = Default_Shape_Square_4x4_Hot_Spot ;
      game->state = Handling_piece ;
      SET_PIECE (game->player_location.handled_piece,
		 0,Orient_Up, Shape_Square_4x4) ;
      clamp_player_location (game) ;
      refresh_shape_count (game, Shape_Square_4x4) ;
      ret_val = 2 ;         /* On fait clignoter la pice au retour. */
    }
    else {
      play_sound (g_wav_no) ;
      shake_and_blink_screen () ;
    }
    break ;                  /* Fin du cas Area_Select_Shape_Square_4x4 */

  case Area_Select_Shape_Long:
    if (game->remaining_pieces[Shape_Long - 1] > 0) {
      game->remaining_pieces[Shape_Long - 1]-- ;
      game->player_location.player_area = Area_board ;
      game->player_location.hot_spot = Default_Shape_Long_Hot_Spot ;
      game->state = Handling_piece ;
      SET_PIECE (game->player_location.handled_piece,
		 0, Orient_Up, Shape_Long) ;
      clamp_player_location (game) ;
      refresh_shape_count (game, Shape_Long) ;
      ret_val = 2 ;         /* On fait clignoter la pice au retour. */
    }
    else {
      play_sound (g_wav_no) ;
      shake_and_blink_screen () ;
    }
    break ;                  /* Fin du cas Area_Select_Shape_Long. */

  case Area_Select_Shape_Junction:
    if (game->remaining_pieces[Shape_Junction - 1] > 0) {
      game->remaining_pieces[Shape_Junction - 1]-- ;
      game->player_location.player_area = Area_board ;
      game->player_location.hot_spot = Default_Shape_Junction_Hot_Spot ;
      game->state = Handling_piece ;
      SET_PIECE (game->player_location.handled_piece,
		 0, Orient_Up, Shape_Junction) ;
      clamp_player_location (game) ;
      refresh_shape_count (game, Shape_Junction) ;
      ret_val = 2 ;         /* On fait clignoter la pice au retour. */
    }
    else {
      play_sound (g_wav_no) ;
      shake_and_blink_screen () ;
    }
    break ;                  /* Fin du cas Area_Select_Shape_Junction. */

  case Area_Select_Shape_Stair_1:
    if (game->remaining_pieces[Shape_Stair_1 - 1] > 0) {
      game->remaining_pieces[Shape_Stair_1 - 1]-- ;
      game->player_location.player_area = Area_board ;
      game->player_location.hot_spot = Default_Shape_Stair_1_Hot_Spot ;
      game->state = Handling_piece ;
      SET_PIECE (game->player_location.handled_piece,
		 0, Orient_Up, Shape_Stair_1) ;
      clamp_player_location (game) ;
      refresh_shape_count (game, Shape_Stair_1) ;
      ret_val = 2 ;         /* On fait clignoter la pice au retour. */
    }
    else {
      play_sound (g_wav_no) ;
      shake_and_blink_screen () ;
    }
    break ;                  /* Fin du cas Area_Select_Shape_Stair_1. */

  case Area_Select_Shape_Stair_2:
     if (game->remaining_pieces[Shape_Stair_2 - 1] > 0) {
      game->remaining_pieces[Shape_Stair_2 - 1]-- ;
      game->player_location.player_area = Area_board ;
      game->player_location.hot_spot = Default_Shape_Stair_2_Hot_Spot ;
      game->state = Handling_piece ;
      SET_PIECE (game->player_location.handled_piece,
		 0, Orient_Up, Shape_Stair_2) ;
      clamp_player_location (game) ;
      refresh_shape_count (game, Shape_Stair_2) ;
      ret_val = 2 ;         /* On fait clignoter la pice au retour. */
    }
    else {
      play_sound (g_wav_no) ;
      shake_and_blink_screen () ;
    }
    break ;                  /* Fin du cas Area_Select_Shape_Stair_1. */

  case Area_Select_Shape_Corner:
     if (game->remaining_pieces[Shape_Corner - 1] > 0) {
      game->remaining_pieces[Shape_Corner - 1]-- ;
      game->player_location.player_area = Area_board ;
      game->player_location.hot_spot = Default_Shape_Corner_Hot_Spot ;
      game->state = Handling_piece ;
      SET_PIECE (game->player_location.handled_piece,
		 0, Orient_Up, Shape_Corner) ;
      clamp_player_location (game) ;
      refresh_shape_count (game, Shape_Corner) ;
      ret_val = 2 ;         /* On fait clignoter la pice au retour. */
    }
    else {
      play_sound (g_wav_no) ;
      shake_and_blink_screen () ;
    }
    break ;                  /* Fin du cas Area_Select_Shape_Corner. */
  }               /* Fin de switch (game->player_location.player_area). */

  return (ret_val) ;
}





enum game_end game_main_loop (struct game *game)
{
  u8 end = 0 ;
  short moved = 0 ;
  short pause_result ;

  start_timer2 () ;
  /* Raffrachissement initial du curseur du joueur. */
  refresh_player_cursor (game) ;
  while (!end) {
    moved = 0 ;

    /* Acquisition des touches. */
    KEY_POOL ;

    if (KEY_HIT (BUTTON_A)) {
      moved |= a_button_pressed (game) ;
      goto refresh_gfxs ;
    }

    if (KEY_HIT (BUTTON_B)) {
      moved |= b_button_pressed (game) ;
      goto refresh_gfxs ;
    }

    if (KEY_HIT (BUTTON_LEFT)) {
      moved |= l_or_r_button_pressed (1, game) ;
      goto refresh_gfxs ;
    }

    if (KEY_HIT (BUTTON_RIGHT)) {
      moved |= l_or_r_button_pressed (0, game) ;
      goto refresh_gfxs ;
    }

    if (KEY_HIT (KEYPAD_DOWN)) {
      moved |= move_cursor_down (game) ;
      goto refresh_gfxs ;
    }

    if (KEY_HIT (KEYPAD_LEFT)) {
      moved |= move_cursor_left (game) ;
      goto refresh_gfxs ;
    }

    if (KEY_HIT (KEYPAD_RIGHT)) {
      moved |= move_cursor_right (game) ;
      goto refresh_gfxs ;
    }

    if (KEY_HIT (KEYPAD_UP)) {
      moved |= move_cursor_up (game) ;
      goto refresh_gfxs ;
    }

    if (KEY_HIT (BUTTON_START)) {
      stop_eyes_sequence () ;
      pause_result = start_button_pressed (game) ;
      if (pause_result == PAUSE_SELECTION_RESTART)
	return (Please_retry) ;    /* Fin force du jeu courant. */
      if (pause_result == PAUSE_SELECTION_ABANDON)
	return (I_give_up) ;
      moved = 1 ;
      goto refresh_gfxs ;
    }

  refresh_gfxs:
    /* Toutes les 12 secondes, on balance les yeux. */
    if ((elapsed_seconds % 12) == 0) start_eyes_sequence () ;
    handle_eyes_animation () ;
    /* On ne raffrachit le pointeur que si l'on a effectivement boug. */
    if (moved) refresh_player_cursor (game) ;
    /* On regarde s'il faut faire clignoter la pice. */
    if (moved == 2) cycle_shape_palette () ;
    end = (game->nb_remaining_cells == 0) ;
  }

  stop_eyes_sequence () ;
  stop_timer2 () ;
  return (Goal_reached) ;
}
