#include <stdlib.h>
#include <string.h>
#include <gba.h>
#include <interrupt.h>
#include "it_handlers.h"
#include <screen_mode.h>
#include "shapes.h"
#include "cells_gfx.h"
#include "large_digits_gfx.h"
#include "shapes_gfx.h"
#include "finger_sprite_gfx.h"
#include "arrow_sprite_gfx.h"
#include "eyes_sprite_gfx.h"
#include "shapes_sprites_gfx.h"





void shake_and_blink_screen ()
{
  u16 old_blending_mode ;
  short i ;

  old_blending_mode = REG_BLDMOD ;
  REG_COLEY = 0 ;
  set_blending_mode (FADE_MODE_2 | FADE_SOURCE_BG1) ;
  for (i = 0; i < 16; i++) {
    while (REG_VCOUNT != 160) ;
    REG_COLEY = (u16) i ;
    REG_BG0HOFS = ((s16) ((rand ()) % 8)) - 4 ;
    REG_BG0VOFS = ((s16) ((rand ()) % 8)) - 4 ;
  }
  for (i = 15; i >= 0; i--) {
    while (REG_VCOUNT != 160) ;
    REG_COLEY = (u16) i ;
    REG_BG0HOFS = ((s16) ((rand ()) % 8)) - 4 ;
    REG_BG0VOFS = ((s16) ((rand ()) % 8)) - 4 ;
  }

  REG_BG0HOFS = 0 ;
  REG_BG0VOFS = 0 ;
  set_blending_mode (old_blending_mode) ;
}


struct eyes_anim_info {
  u32 time ;
  u8 running ;
  u8 nb_frames ;
  u8 frame_counter ;
} ;


static struct eyes_anim_info anim_info = { 0, 0, 0, 0 } ;
static struct sprite eyes_sprite  = { 0, 0, 0, 0 } ;


void start_eyes_sequence ()
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;

  anim_info.time = elapsed_seconds ;
  anim_info.running = 1 ;
  anim_info.nb_frames = 3 + ((rand ()) % 7) ;
  anim_info.frame_counter = 0 ;

  eyes_sprite.attribute0 =
    SPRITE_16_COLOR | SPRITE_WIDE | 0 /* y */ ;
  eyes_sprite.attribute1 = SPRITE_SIZE_32 | 0 /* x */ ;
  eyes_sprite.attribute2 =
    EYES_SPRITE_LEFT_TILE_INDEX | (EYES_SPRITE_PALETTE_SLOT << 12) ;

  while (REG_VCOUNT < 160) ;
  memcpy (&sprite_video_memory[RESERVED_EYES_SPRITE_INDEX], &eyes_sprite,
	  sizeof (struct sprite)) ;
}



void stop_eyes_sequence ()
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;

  anim_info.running = 0 ;
  eyes_sprite.attribute0 = 160 ; /* y vers > 159. */
  eyes_sprite.attribute1 = 240 ; /* x vers  > 239. */
  while (REG_VCOUNT < 160) ;
  memcpy (&sprite_video_memory[RESERVED_EYES_SPRITE_INDEX], &eyes_sprite,
	  sizeof (struct sprite)) ;
}



void handle_eyes_animation ()
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;

  if (anim_info.running == 0) return ;
  if (elapsed_seconds <= anim_info.time) return ;

  anim_info.frame_counter++ ;

  /* Fin de l'animation ? */
  if (anim_info.frame_counter >= anim_info.nb_frames) {
    stop_eyes_sequence () ;
    return ;
  }

  /* Non ! */
  switch (anim_info.frame_counter % 4) {
  case 0:
    eyes_sprite.attribute2 =
      EYES_SPRITE_LEFT_TILE_INDEX | (EYES_SPRITE_PALETTE_SLOT << 12) ;
    break ;

  case 1:
    eyes_sprite.attribute2 =
      EYES_SPRITE_CENTER_TILE_INDEX | (EYES_SPRITE_PALETTE_SLOT << 12) ;
    break ;

  case 2:
    eyes_sprite.attribute2 =
      EYES_SPRITE_RIGHT_TILE_INDEX | (EYES_SPRITE_PALETTE_SLOT << 12) ;
    break ;

  default:
    eyes_sprite.attribute2 =
      EYES_SPRITE_CENTER_TILE_INDEX | (EYES_SPRITE_PALETTE_SLOT << 12) ;
    break ;
  }

  while (REG_VCOUNT < 160) ;
  memcpy (&sprite_video_memory[RESERVED_EYES_SPRITE_INDEX], &eyes_sprite,
	  sizeof (struct sprite)) ;

  anim_info.time = elapsed_seconds ;
}





void cycle_shape_palette ()
{
  u16 i, j, turn ;
  const u16 *palette ;
  u16 *sprite_palette = SPRITES_PALETTE_ADR ;

  for (turn = 0; turn < 15; turn++) {
    i = SHAPES_SPRITES_PALETTE_SLOT * 16 ;
    if (turn % 2) palette = shapes_palette ;
    else palette = shapes_alternate_palette ;
    while (REG_VCOUNT != 160) ;
    for (j = 0; j < SHAPES_PALETTE_SIZE; i++, j++)
      sprite_palette[i] = palette[j] ;
  }

  /* Restoration de la palette initiale. */
  i = SHAPES_SPRITES_PALETTE_SLOT * 16 ;
  while (REG_VCOUNT != 160) ;
  for (j = 0; j < SHAPES_PALETTE_SIZE; i++, j++)
    sprite_palette[i] = shapes_palette[j] ;
}
