#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gba.h>
#include <screen_mode.h>
#include "shapes.h"
#include "load_gfx_data.h"
#include "layout_gfx.h"
#include "cells_gfx.h"
#include "large_digits_gfx.h"
#include "shapes_gfx.h"
#include "finger_sprite_gfx.h"
#include "game_core.h"
#include "ihm.h"
#include "bg1_tiles_layout.h"
#include "arrow_sprite_gfx.h"
#include "eyes_sprite_gfx.h"
#include "shapes_sprites_gfx.h"
#include "shapes_configurations.h"
#include <interrupt.h>
#include "it_handlers.h"





void init_persistent_sprites_data ()
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;
  short i ;

  /* Chargement des sprites reprsentant les diffrents curseurs du joueur. */
  /* D'abord le sprite reprsentant la flche. */
  load_sprites_16 (ARROW_SPRITE_LOADING_OFFSET,   /* Offset de chargement. */
		   ARROW_SPRITE_NB_TILES,         /* Nb tiles. */
		   arrow_sprite_tiles,            /* Tiles src. */
		   ARROW_SPRITE_PALETTE_SLOT,     /* Slot palette. */
		   ARROW_SPRITE_PALETTE_SIZE,     /* Taille palette. */
		   arrow_sprite_palette) ;        /* Palette src. */
  /* Ensuite le sprite reprsentant le doigt. */
  load_sprites_16 (FINGER_SPRITE_LOADING_OFFSET,   /* Offset de chargement. */
		   FINGER_SPRITE_NB_TILES,         /* Nb tiles. */
		   finger_sprite_tiles,            /* Tiles src. */
		   FINGER_SPRITE_PALETTE_SLOT,     /* Slot palette. */
		   FINGER_SPRITE_PALETTE_SIZE,     /* Taille palette. */
		   finger_sprite_palette) ;        /* Palette src. */
  /* Ensuite on charge les sprites d'anim des yeux. */
  load_sprites_16 (EYES_SPRITE_LOADING_OFFSET,     /* Offset de chargement. */
		   EYES_SPRITE_NB_TILES,           /* Nb tiles. */
		   eyes_sprite_tiles,              /* Tiles src. */
		   EYES_SPRITE_PALETTE_SLOT,       /* Slot palette. */
		   EYES_SPRITE_PALETTE_SIZE,       /* Taille palette. */
		   eyes_sprite_palette) ;          /* Palette src. */
  /* Ensuite, on ne charge pas les tiles de sprites reprsentant les */
  /* formes, mais seulement la palette utilise par ces sprites.     */
  load_sprites_16 (SHAPES_SPRITES_LOADING_OFFSET,   /* Offset de chargement. */
		   0,                               /* Nb tiles. */
		   NULL,                            /* Tiles src. */
		   SHAPES_SPRITES_PALETTE_SLOT,     /* Slot palette. */
		   SHAPES_PALETTE_SIZE,             /* Taille palette. */
		   shapes_palette) ;                /* Palette src. */

  /* Cache tous les sprites en dehors de l'cran pour que ceux que l'on */
  /* n'utilisera pas ne se voient pas et on vide le reste de la mmoire */
  /* contenant les attributs des sprites.                               */
  while (REG_VCOUNT < 160) ;
  for (i = 0; i < 128; i++) {
    sprite_video_memory[i].attribute0 = 160 ; /* y vers > 159. */
    sprite_video_memory[i].attribute1 = 240 ; /* x vers  > 239. */
    sprite_video_memory[i].attribute2 = 0 ;
    sprite_video_memory[i].attribute3 = 0 ;
  }
}





void setup_board_graphics ()
{
  unsigned short i ;

  set_screen_mode (0) ;

  /* Initialisations pour le background 0 (layout de base). */
  set_background0
    (BACKGROUND_PRIORITY_3 |            /* Priorit la plus faible. */
     BACKGROUND_CHAR_BASE_BLOCK_0 |     /* Tiles */
     BACKGROUND_COLOR_256 |             /* 256 couleurs. */
     BACKGROUND_SCREEN_BASE_BLOCK_8 |   /* Map. */
     TEXT_BACKGROUND_256x256) ;         /* 256 x 256 pixels. */

  /* Chargement des tiles et de la map de fond d'cran de jeu. */
  load_palette_tile_map_256
    (CHAR_BASE_BLOCK_0_ADDR,       /* Tiles dest. */
     SCREEN_BASE_BLOCK_8_ADDR,     /* Map dest. */
     LAYOUT_PALETTE_SIZE,          /* Taille palette. */
     layout_palette,               /* Palette src. */
     LAYOUT_NB_TILES,              /* Nb tiles. */
     layout_tiles,                 /* Tiles src. */
     LAYOUT_MAP_WIDTH,             /* Largeur map. */
     LAYOUT_MAP_HEIGHT,            /* Hauteur map. */
     layout_map) ;                 /* Map src. */

  /* Initialisations pour le background 1 (cran de jeu dynamique). */
  set_background1
    (BACKGROUND_PRIORITY_2 |            /* Priorit faible. */
     BACKGROUND_CHAR_BASE_BLOCK_2 |     /* Tiles */
     BACKGROUND_COLOR_16 |             /* 16 couleurs. */
     BACKGROUND_SCREEN_BASE_BLOCK_9 |   /* Map. */
     TEXT_BACKGROUND_256x256) ;         /* 256 x 256 pixels. */

  /* Chargement des tiles d'autorisation/interdiction de cellule. */
  /* Pas de map, pas de palette. */
  load_palette_tile_map_16
    (CHAR_BASE_BLOCK_2_ADDR,  /* Tiles dest. */
     NULL,                    /* Map dest. */
     0,                       /* Slot palette. */
     0,                       /* Taille palette. */
     NULL,                    /* Palette src. */
     CELLS_NB_TILES,          /* Nb tiles. */
     cells_tiles,             /* Tiles src. */
     0,                       /* Largeur map. */
     0,                       /* Hauteur map. */
     NULL) ;                  /* Map src. */

  /* Chargement des tiles de grands digits pour le nombre de pices dispos. */
  /* Pas de map, pas de palette. */
  load_palette_tile_map_16
    (/* Tiles dest. On divise par 2 car les tiles sont des u8 */
     /* mais le pointeur sur la mmoire vid est en u16.      */
     CHAR_BASE_BLOCK_2_ADDR + ((CELLS_NB_TILES * 32) / 2),
     NULL,                    /* Map dest. */
     0,                       /* Slot palette. */
     0,                       /* Taille palette. */
     NULL,                    /* Palette src. */
     LARGE_DIGITS_NB_TILES,   /* Nb tiles. */
     large_digits_tiles,      /* Tiles src. */
     0,                       /* Largeur map. */
     0,                       /* Hauteur map. */
     NULL) ;                  /* Map src. */

  /* Chargement des tiles et de la palette des pices.  */
  /* La map vido n'est pas charge dans la mmoire car */
  /* on la construira dynamiquement  partir de la map  */
  /* des pices en ROM.                                 */
  load_palette_tile_map_16
    (/* Tiles dest. */
     CHAR_BASE_BLOCK_2_ADDR + ((CELLS_NB_TILES * 32) / 2)
                            + ((LARGE_DIGITS_NB_TILES * 32) / 2) ,
     NULL,                     /* Map dest. */
     SHAPES_PALETTE_SLOT,      /* Slot palette. */
     SHAPES_PALETTE_SIZE,      /* Taille palette. */
     shapes_palette,           /* Palette src. */
     SHAPES_NB_TILES,          /* Nb tiles. */
     shapes_tiles,             /* Tiles src. */
     0,                        /* Largeur map. */
     0,                        /* Hauteur map. */
     NULL) ;                   /* Map src. */

  /* Effacement du fond du layer BG1 (256x256 pixels => nombre d'u16) . */
  for (i = 0; i < ((256 / 8) * (256 / 8)); i++)
    SCREEN_BASE_BLOCK_9_ADDR[i] =
      (u16) (CELL_EMPTY_INDEX | (SHAPES_PALETTE_SLOT << 12)) ;

  init_persistent_sprites_data () ;

  REG_BG0HOFS = 0 ;
  REG_BG0VOFS = 0 ;
  REG_BG1HOFS = 0 ;
  REG_BG1VOFS = 0 ;
  set_screen_mode
    (MODE_0 | BG0_ENABLE | BG1_ENABLE | OBJ_MAP_1D | OBJ_ENABLE) ;
}




void restore_game_palette ()
{
  load_palette_tile_map_256
    (CHAR_BASE_BLOCK_0_ADDR,       /* Tiles dest. */
     SCREEN_BASE_BLOCK_8_ADDR,     /* Map dest. */
     LAYOUT_PALETTE_SIZE,          /* Taille palette. */
     layout_palette,               /* Palette src. */
     LAYOUT_NB_TILES,              /* Nb tiles. */
     layout_tiles,                 /* Tiles src. */
     LAYOUT_MAP_WIDTH,             /* Largeur map. */
     LAYOUT_MAP_HEIGHT,            /* Hauteur map. */
     layout_map) ;                 /* Map src. */
}









static void large_print_int (u16 col, u16 line, int value)
{
  short column ;
  unsigned int div ;
  unsigned int remain = 0 ;
  u16 tile_on_screen ;
  u16 *output_map = SCREEN_BASE_BLOCK_9_ADDR ;

  div = value ;
  /* Affichage sur 2 chiffres. */
  for (column = 1; column >= 0; column --) {
    remain = div % 10 ;
    div = div / 10 ;
    tile_on_screen = (line * (256 / 8)) + (col + column) ;
    /* Premire tile du chiffre. */
    output_map[tile_on_screen] =
      (u16) ((LARGE_DIGITS_START_INDEX + (remain * 2)) |
	     (SHAPES_PALETTE_SLOT << 12)) ;
    /* Seconde tile du chiffre. */
    tile_on_screen = ((line + 1) * (256 / 8)) + (col + column) ;
    output_map[tile_on_screen] =
      (u16) ((LARGE_DIGITS_START_INDEX + ((remain * 2) + 1)) |
	     (SHAPES_PALETTE_SLOT << 12)) ;
  }
}





void refresh_shape_count (struct game *game, enum shape shape)
{
  short value ;
  u16 l = 0 ;
  u16 c = 0 ;

  switch (shape) {
  case Shape_None: break ;

  case Shape_Square_1x1:
    l = Area_Remaining_Shape_Square_1x1_Y ;
    c = Area_Remaining_Shape_Square_1x1_X ;
    break ;

  case Shape_Square_4x4:
    l = Area_Remaining_Shape_Square_4x4_Y ;
    c = Area_Remaining_Shape_Square_4x4_X ;
    break ;

  case Shape_Long:
    l = Area_Remaining_Shape_Long_Y ;
    c = Area_Remaining_Shape_Long_X ;
    break ;

  case Shape_Junction:
    l = Area_Remaining_Shape_Junction_Y ;
    c = Area_Remaining_Shape_Junction_X ;
    break ;

  case Shape_Stair_1:
    l = Area_Remaining_Shape_Stair_1_Y ;
    c = Area_Remaining_Shape_Stair_1_X ;
    break ;

  case Shape_Stair_2:
    l = Area_Remaining_Shape_Stair_2_Y ;
    c = Area_Remaining_Shape_Stair_2_X ;
    break ;

  case Shape_Corner:
    l = Area_Remaining_Shape_Corner_Y ;
    c = Area_Remaining_Shape_Corner_X ;
    break ;
  }

  value = game->remaining_pieces[shape - 1] ;
  large_print_int (c, l, value) ;
}






/* 4 sprites ncessaires au maximum. */
#define MAX_NUM_PLAYER_SPRITES 4
static struct sprite player_sprites[MAX_NUM_PLAYER_SPRITES] = {
  { 0, 0, 0, 0 },
  { 0, 0, 0, 0 },
  { 0, 0, 0, 0 },
  { 0, 0, 0, 0 }} ;
static short nb_used_player_sprites = 0 ;




void hide_player_sprites ()
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;
  short i ;

  for (i = 0; i < MAX_NUM_PLAYER_SPRITES; i++) {
    /* Effacement des anciens X. */
    player_sprites[i].attribute1 = player_sprites[i].attribute1 & 0xFE00 ;
    /* Effacement de l'ancien Y. */
    player_sprites[i].attribute0 = player_sprites[i].attribute0 & 0xFF00 ;
  }

  nb_used_player_sprites = 0 ;

  /* Attente synchro VBL. */
  while (REG_VCOUNT < 160) ;
  /* Transfert de toutes les donnes sprite "curseurs" en mmoire. */
  memcpy (&sprite_video_memory[0], player_sprites,
	  MAX_NUM_PLAYER_SPRITES * sizeof (struct sprite)) ;
}




void shake_player_shape (struct game *game)
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;
  u16 times, i ;
  u16 x, y ;

  if (nb_used_player_sprites == 0) return ;
  
  for (times = 0; times < 10; times++) {
    for (i = 0; i < nb_used_player_sprites; i++) {
      /* Rcupration des anciennes coordonnes du sprite. */
      x = player_sprites[i].attribute1 & 0x01FF ;
      y = player_sprites[i].attribute0 & 0x00FF ;
      /* Effacement des anciens X. */
      player_sprites[i].attribute1 = player_sprites[i].attribute1 & 0xFE00 ;
      /* Effacement de l'ancien Y. */
      player_sprites[i].attribute0 = player_sprites[i].attribute0 & 0xFF00 ;
      x += (4 - ((s16) ((rand ()) % 8))) ;
      y += (4 - ((s16) ((rand ()) % 8))) ;
      player_sprites[i].attribute1 = player_sprites[i].attribute1 | x ;
      player_sprites[i].attribute0 = player_sprites[i].attribute0 | y ;
    }

    while (REG_VCOUNT != 160) ;
    /* Transfert de toutes les donnes sprite "curseurs" en mmoire. */
    memcpy (&sprite_video_memory[0], player_sprites,
	    nb_used_player_sprites * sizeof (struct sprite)) ;
  }

  /* Finalement, restauration du curseur comme il faut. */
  refresh_player_cursor (game) ;
}




typedef s16 Fixed ;
#define TO_FIXED(v) ((Fixed) ((v) * (1<<8)))

void mag_min_player_shape (struct game *game)
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;
  u16 times, i ;
  Fixed scaling ;

  if (nb_used_player_sprites == 0) return ;

  /* On autorise le scaling. */
  for (i = 0; i < nb_used_player_sprites; i++) {
    player_sprites[i].attribute0 |= SPRITE_ROTATION_FLAG ;
    /* Index de scaling: 0 */
    player_sprites[i].attribute1 =
      (player_sprites[i].attribute1 & (~ 0x3E00)) | (0 << 9) ;
  }

  for (times = 0; times < 10; times++) {
    scaling = TO_FIXED ((1.0 / ((float) ((times % 2) + 1)))) ;
    player_sprites[0].attribute3 = scaling ;
    player_sprites[1].attribute3 = 0 ;
    player_sprites[2].attribute3 = 0 ;
    player_sprites[3].attribute3 = scaling ;
    
    while (REG_VCOUNT != 160) ;
    /* Transfert de toutes les donnes sprite "curseurs" en mmoire. */
    memcpy (&sprite_video_memory[0], player_sprites,
	    nb_used_player_sprites * sizeof (struct sprite)) ;
  }

  /* Finalement, restauration du curseur comme il faut. */
  refresh_player_cursor (game) ;
}




void refresh_player_cursor (struct game *game)
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;
  enum orientation tmp_orient ;
  enum shape tmp_shape ;
  struct chain_code *tmp_ccode = NULL ; ;
  const u8 *tmp_map ;
  short i ;
  short tmp_x, tmp_y ;

  for (i = 0; i < MAX_NUM_PLAYER_SPRITES; i++) {
    /* Effacement des anciens X. */
    player_sprites[i].attribute1 = player_sprites[i].attribute1 & 0xFE00 ;
    /* Effacement de l'ancien Y. */
    player_sprites[i].attribute0 = player_sprites[i].attribute0 & 0xFF00 ;
  }

  switch (game->player_location.player_area) {
  case Area_board:
    /* Il faut soit dessiner la pice tenue, soit s'il n'y en n'a pas, */
    /* dessiner le pointeur permettant de dsigner la pice  retirer. */
    switch (game->state) {
    case Free_hand:
      /* Il faut dessiner le pointeur permettant de choisir */
      /* une pice dans le pool des pices disponibles.     */
      nb_used_player_sprites = 1 ;
      /* On positionne y en mme temps. */
      player_sprites[0].attribute0 =
	SPRITE_16_COLOR | SPRITE_SQUARE |
	((Area_board_TOP_Y * 8) + (game->player_location.y_as_index * 8)) ;
      /* On positionne x en mme temps. */
      player_sprites[0].attribute1 =
	SPRITE_SIZE_32 |
	((Area_board_TOP_X * 8) + (game->player_location.x_as_index * 8)) ;
      player_sprites[0].attribute2 =
	FINGER_SPRITE_TILE_INDEX | (FINGER_SPRITE_PALETTE_SLOT << 12) ;
      break ;

    case Handling_piece:
      tmp_orient = GET_ORIENT (game->player_location.handled_piece) ;
      tmp_shape = GET_SHAPE (game->player_location.handled_piece) ;
      if (tmp_shape == Shape_None) {
	// TODO ASSERT FALSE.
      }
      else {
	/* On dcrmente le code de la forme car 0 est utilis pour   */
	/* dire "pas de pice" or tous les tableaux de configurations */
	/* sont indexs sur le code des pices, mais  partir de 0.   */
	tmp_shape-- ;
	tmp_ccode =
	  (struct chain_code*) rotated_shapes_ccodes[tmp_shape][tmp_orient] ;
	tmp_map = rotated_shapes_maps[tmp_shape][tmp_orient] ;
	/* Nombre de sprites ncessaires. */
	nb_used_player_sprites = tmp_ccode->length ;
	/* On fait en sorte que la pice se trouve dans la zone  */
	/* de jeu au plus prs de la position du joueur.         */
	tmp_x =
	  (Area_board_TOP_X * 8) + (game->player_location.x_as_index * 8) ;
	tmp_y =
	  (Area_board_TOP_Y * 8) + (game->player_location.y_as_index * 8) ;
	for (i = 0; i < nb_used_player_sprites; i++) {
	  /* Ordonne. */
	  player_sprites[i].attribute0 =
	    SPRITE_16_COLOR | SPRITE_SQUARE | tmp_y ;
	  /* Abscisse et taille. */
	  player_sprites[i].attribute1 = SPRITE_SIZE_8 | tmp_x ;
	  /* Chargement des donnes graphiques de la tile du sprite. */
	  load_sprites_16
	    (SHAPES_SPRITES_LOADING_OFFSET + ((i * 32) / 2),
	     1, &shapes_tiles[32 * tmp_map[i]],
	     SHAPES_SPRITES_PALETTE_SLOT, 0, NULL) ;
	  /* Tile et palette. */
	  player_sprites[i].attribute2 =
	    (SHAPES_SPRITES_TILES_INDEX + i) |
	    (SHAPES_SPRITES_PALETTE_SLOT << 12) ;
	  tmp_x += (8 * tmp_ccode->codes[i].x_delta) ;
	  tmp_y += (8 * tmp_ccode->codes[i].y_delta) ;
	}
      }    /* Fin du else if (GET_SHAPE(game->player_location.handled_ ... */
      break ;                     /* Fin du cas Handling_piece. */
    }                             /* Fin du "switch (game->game_state)". */
    break ;                       /* Fin du cas "Area_board". */

  case Area_Select_Shape_Square_1x1:
    player_sprites[0].attribute1 =
      player_sprites[0].attribute1 | (Area_Select_Shape_Square_1x1_X * 8) ;
    player_sprites[0].attribute0 =
      player_sprites[0].attribute0 | (Area_Select_Shape_Square_1x1_Y * 8) ;
    goto set_common_tile_index ;
    /* False through !!! */

  case Area_Select_Shape_Square_4x4:
    player_sprites[0].attribute1 =
      player_sprites[0].attribute1 | (Area_Select_Shape_Square_4x4_X * 8) ;
    player_sprites[0].attribute0 =
      player_sprites[0].attribute0 | (Area_Select_Shape_Square_4x4_Y * 8) ;
    goto set_common_tile_index ;
    /* False through !!! */

  case Area_Select_Shape_Long:
    player_sprites[0].attribute1 =
      player_sprites[0].attribute1 | (Area_Select_Shape_Long_X * 8) ;
    player_sprites[0].attribute0 =
      player_sprites[0].attribute0 | (Area_Select_Shape_Long_Y * 8) ;
    goto set_common_tile_index ;
    /* False through !!! */

  case Area_Select_Shape_Junction:
    player_sprites[0].attribute1 =
      player_sprites[0].attribute1 | (Area_Select_Shape_Junction_X * 8) ;
    player_sprites[0].attribute0 =
      player_sprites[0].attribute0 | (Area_Select_Shape_Junction_Y * 8) ;
    goto set_common_tile_index ;
    /* False through !!! */

  case Area_Select_Shape_Stair_1:
    player_sprites[0].attribute1 =
      player_sprites[0].attribute1 | (Area_Select_Shape_Stair_1_X * 8) ;
    player_sprites[0].attribute0 =
      player_sprites[0].attribute0 | (Area_Select_Shape_Stair_1_Y * 8) ;
    goto set_common_tile_index ;
    /* False through !!! */

  case Area_Select_Shape_Stair_2:
    player_sprites[0].attribute1 =
      player_sprites[0].attribute1 | (Area_Select_Shape_Stair_2_X * 8) ;
    player_sprites[0].attribute0 =
      player_sprites[0].attribute0 | (Area_Select_Shape_Stair_2_Y * 8) ;
    goto set_common_tile_index ;
    /* False through !!! */

  case Area_Select_Shape_Corner:
    player_sprites[0].attribute1 =
      player_sprites[0].attribute1 | (Area_Select_Shape_Corner_X * 8) ;
    player_sprites[0].attribute0 =
      player_sprites[0].attribute0 | (Area_Select_Shape_Corner_Y * 8) ;
    /* False through !!! */

  set_common_tile_index:
    /* Il faut dessiner le pointeur permettant de choisir */
    /* une pice dans le pool des pices disponibles.     */
    nb_used_player_sprites = 1 ;
    player_sprites[0].attribute0 |= SPRITE_16_COLOR | SPRITE_SQUARE ;
    player_sprites[0].attribute1 |= SPRITE_SIZE_32 ;
    player_sprites[0].attribute2 =
      FINGER_SPRITE_TILE_INDEX | (FINGER_SPRITE_PALETTE_SLOT << 12) ;
    break ;              /* Fin de tous les cas "Area-de-pices". */
  }                      /* Fin du "(game->player_location.player_area)". */

  /* Cache tous les sprites "curseurs" non utiliss. */
  for (i = nb_used_player_sprites; i < MAX_NUM_PLAYER_SPRITES; i++) {
    player_sprites[i].attribute0 = 160 ; /* y vers > 159. */
    player_sprites[i].attribute1 = 240 ; /* x vers  > 239. */
  }

  /* Attente synchro VBL. */
  while (REG_VCOUNT < 160) ;
  /* Transfert de toutes les donnes sprite "curseurs" en mmoire. */
  memcpy (&sprite_video_memory[0], player_sprites,
	  MAX_NUM_PLAYER_SPRITES * sizeof (struct sprite)) ;
}




void initial_board_refresh (struct game *game)
{
  short i, x, y ;
  u16 *output_map = SCREEN_BASE_BLOCK_9_ADDR ;
  u16 tile_on_screen ;

  /* Dessin de la forme du but. */
  for (y = 0; y < BOARD_HEIGHT; y++) {
    tile_on_screen = ((y + Area_board_TOP_Y) * (256/8)) + Area_board_TOP_X ;
    for (x = 0; x < BOARD_WIDTH; x++) {
      if (game->aim[x][y] == 1)
	output_map[tile_on_screen] =
	  (u16) (CELL_EMPTY_INDEX | (SHAPES_PALETTE_SLOT << 12)) ;
      else
	output_map[tile_on_screen] =
	  (u16) (CELL_FORBIDDEN_INDEX | (SHAPES_PALETTE_SLOT << 12)) ;
      tile_on_screen ++ ;
    }
  }

  /* Affichage initial des nombres de pices  disposition. */
  for (i = 0; i < SHAPES_NUMBER; i++)
    refresh_shape_count (game, ((enum shape) (i + 1))) ;
}
