#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gba.h>
#include <screen_mode.h>
#include "shapes.h"
#include "key_control.h"
#include "game_core.h"
#include "load_gfx_data.h"
#include "lev_sel_gfx.h"
#include "lev_sel.h"
#include "arrow_sprite_gfx.h"
#include "finger_sprite_gfx.h"
#include "ihm.h"
#include "levels_data.h"
#include "persistent.h"



static struct sprite finger_sprite = { 0, 0, 0, 0 } ;
#define ON_COLOR 234
#define OFF_COLOR_DONE 178
#define OFF_COLOR_TODO 188
#define NONE_COLOR 1

#define INLINED_DEFLATE \
    for (bit = 0; bit < 8; bit++) {\
      if (data & 0x01) *dest = ON_COLOR ;\
      else *dest = off_color ;\
      data = data >> 1 ;\
      dest-- ;\
    }


/* Si on le met en local, apparemment a fait pter la pile. */
static u8 tile_buffer[4][64] ;

static void deflate_snapshot (u8 *tile_start_addr, u16 *compressed_data,
			      u8 done_flag)
{
  short line, bit ;
  u8 data ;
  u8 *dest ;
  u16 off_color ;
  
  /* Couleur de fond dpendant de si le niveau a dj t rsolu. */
  off_color = (done_flag ? OFF_COLOR_DONE : OFF_COLOR_TODO) ;
  if (compressed_data != NULL) {
    /* Les 8 premires lignes adressent les blocs de donnes de tile 0 et 1. */
    for (line = 0; line < 8; line++) {
      /* L'octet de poid fort adresse le bloc 0. */
      dest = tile_buffer[0] + ((line + 1) * 8) - 1 ;
      data = (compressed_data[line] & 0xFF00) >> 8 ;
      INLINED_DEFLATE

      /* L'octet de poid faible adresse le bloc 1. */
      dest = tile_buffer[1] + ((line + 1) * 8) - 1 ;
      data = compressed_data[line] & 0x00FF ;
      INLINED_DEFLATE
    }

    /* Les 8 dernires lignes adressent les blocs de donnes tile 2 et 3. */
    for (line = 0;line < 8; line++) {
      /* L'octet de poid fort adresse le bloc 2. */
      dest = tile_buffer[2] + ((line + 1) * 8) - 1 ;
      data = (compressed_data[line + 8] & 0xFF00) >> 8 ;
      INLINED_DEFLATE

      /* L'octet de poid faible adresse le bloc 3. */
      dest = tile_buffer[3] + ((line + 1) * 8) - 1 ;
      data = compressed_data[line + 8] & 0x00FF ;
      INLINED_DEFLATE
    }
  }
  else {
    /* Pas donnes  dcompresser. On remplit avec la couleur OFF. */
    for (line = 0; line < 4; line++) {
      for (bit = 0; bit < 64; bit++)
	tile_buffer[line][bit] = NONE_COLOR ;
    }
  }

  /* Maintenant, on copie les 4 tiles bufferises en mmoire relle. */
  load_palette_tile_map_256
    ((u16*) tile_start_addr, NULL, 0, NULL,4, (u8*) tile_buffer, 0, 0, NULL) ;
}





static void display_snapshots (int start_level_offset)
{
  int level ;
  u16 x, y ;
  u16 tile_index = LEV_SEL_NB_TILES ;
  u16 *map = SCREEN_BASE_BLOCK_9_ADDR ;
  u8 *tiles = ((u8*) CHAR_BASE_BLOCK_0_ADDR) + (LEV_SEL_NB_TILES * 64) ;
  u16 *compressed_data ;

  for (level = 0;level < 20; level++) {
    if ((start_level_offset + level) >= NB_AVAILABLE_LEVELS)
      compressed_data = NULL ;
    else
      compressed_data = (u16*) levels[start_level_offset + level].bits ;
    /* On dcompresse. */
    deflate_snapshot
      (tiles, compressed_data, progression[start_level_offset + level]) ;
    
    /* Calcul des offset de dbut de la map reprsentant ce snapshot. */
    x = ((level % 7) * Selection_Lev_X_Delta) + Selection_Lev_X_Offset ;
    y = ((level / 7) * Selection_Lev_Y_Delta) + Selection_Lev_Y_Offset ;
    /* On met  jour les 4 entres concernes de la map. */
    map[(y * (256 / 8)) + x] = tile_index ;
    map[(y * (256 / 8)) + x + 1] = tile_index + 1 ;
    map[((y + 1) * (256 / 8)) + x] = tile_index + 2 ;
    map[((y + 1) * (256 / 8)) + x + 1] = tile_index + 3 ;
    /* On passe 4 blocs de donnes de tile plus loin. */
    tiles += 4 * 64 ;
    tile_index += 4 ;
  }
}





int handle_level_input (int start_level_offset)
{
  u16 x = 0, y = 0 ;
  int start_display ;
  enum level_selection cursor_selection, old_cursor_selection ;
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;  

  set_screen_mode (0) ;

  /* Initialisations pour le background 0 (layout de base). */
  set_background0
    (BACKGROUND_PRIORITY_3 |            /* Priorit la plus faible. */
     BACKGROUND_CHAR_BASE_BLOCK_0 |     /* Tiles */
     BACKGROUND_COLOR_256 |             /* 256 couleurs. */
     BACKGROUND_SCREEN_BASE_BLOCK_9 |   /* Map. */
     TEXT_BACKGROUND_256x256) ;         /* 256 x 256 pixels. */

  /* Pas besoin de charger la palette, les tiles ni la map. */
  /* Elles sont dj mises en mmoire.                      */
  /* Par contre, il faut gnrer les snapshots des puzzles. */
  /* Il y a vingt niveaux proposs  chaque page.           */
  start_display = start_level_offset - (start_level_offset % 20) ;
  display_snapshots (start_display) ;
  set_screen_mode (MODE_0 | BG0_ENABLE | OBJ_MAP_1D | OBJ_ENABLE) ;

  /* On se positionne directement sur le niveau point. */
  cursor_selection = (enum level_selection) (start_level_offset % 20) ;
  x = (((cursor_selection % 7)* Selection_Lev_X_Delta) +
       Selection_Lev_X_Offset) * 8 ;
  y = (((cursor_selection / 7) * Selection_Lev_Y_Delta) +
       Selection_Lev_Y_Offset) * 8 + Finger_Sprite_Y_Correction ;
  finger_sprite.attribute0 = SPRITE_16_COLOR | SPRITE_SQUARE ;
  finger_sprite.attribute1 = SPRITE_SIZE_32 ;
  finger_sprite.attribute2 =
    FINGER_SPRITE_TILE_INDEX | (FINGER_SPRITE_PALETTE_SLOT << 12) ;
  finger_sprite.attribute1 = finger_sprite.attribute1 | x ;
  finger_sprite.attribute0 = finger_sprite.attribute0 | y ;
  while (REG_VCOUNT < 160) ;
  memcpy (&sprite_video_memory[0], &finger_sprite, sizeof (struct sprite)) ;

  for (;;) {
    old_cursor_selection = cursor_selection ;
    KEY_POOL ;

    if (KEY_HIT (BUTTON_A)) {
      if (cursor_selection == LEV_SELECTION_BACK) return (-1) ;
      if ((start_display + cursor_selection) < NB_AVAILABLE_LEVELS)
	return (start_display + cursor_selection) ;
    }

    if (KEY_HIT (BUTTON_B)) {
      /* On retourne un indice de niveau impossible */
      /* pour signifier la non-slection.           */
      return (-1) ;
    }

    if (KEY_HIT (KEYPAD_UP)) {
      /* Dplacement vers le haut. */
      cursor_selection = (cursor_selection + 21 - 7) % 21 ;
    }

    if (KEY_HIT (KEYPAD_DOWN)) {
      /* Dplacement vers le bas. */
      cursor_selection = (cursor_selection + 7) % 21 ;
    }

    if (KEY_HIT (KEYPAD_LEFT)) {
      /* Dplacement vers la gauche. */
      cursor_selection = (cursor_selection + 21 - 1) % 21 ;
    }

    if (KEY_HIT (KEYPAD_RIGHT)) {
      /* Dplacement vers la droite. */
      cursor_selection = (cursor_selection + 1) % 21 ;
    }

    if (cursor_selection != old_cursor_selection) {
      u16 x = 0, y = 0 ;

      x = (((cursor_selection % 7)* Selection_Lev_X_Delta) +
	   Selection_Lev_X_Offset) * 8 ;
      y = (((cursor_selection / 7) * Selection_Lev_Y_Delta) +
	   Selection_Lev_Y_Offset) * 8 + Finger_Sprite_Y_Correction ;

      finger_sprite.attribute1 = finger_sprite.attribute1 & 0xFE00 ;
      finger_sprite.attribute1 = finger_sprite.attribute1 | x ;
      finger_sprite.attribute0 = finger_sprite.attribute0 & 0xFF00 ;
      finger_sprite.attribute0 = finger_sprite.attribute0 | y ;  /* y */
      /* Transfert des infos de sprite dans la mmoire vido-sprite. */
      while (REG_VCOUNT < 160) ;
      memcpy (sprite_video_memory, &finger_sprite, sizeof (struct sprite)) ;
    }
  }
}





static int handle_directory_input (int pre_selected_level)
{
  enum directory_selection cursor_selection, old_cursor_selection ;
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;  
  int level_start = 0 ;

  /* Oh que c'est pas propre :)) */
 dirty_re_init:
  /* Initialisations pour le background 0 (layout de base). */
  set_background0
    (BACKGROUND_PRIORITY_3 |            /* Priorit la plus faible. */
     BACKGROUND_CHAR_BASE_BLOCK_0 |     /* Tiles */
     BACKGROUND_COLOR_256 |             /* 256 couleurs. */
     BACKGROUND_SCREEN_BASE_BLOCK_8 |   /* Map. */
     TEXT_BACKGROUND_256x256) ;         /* 256 x 256 pixels. */

  cursor_selection = DIR_SELECTION_1_20 ;
  finger_sprite.attribute0 = SPRITE_16_COLOR | SPRITE_SQUARE ;
  finger_sprite.attribute1 = SPRITE_SIZE_32 ;
  finger_sprite.attribute2 =
    FINGER_SPRITE_TILE_INDEX | (FINGER_SPRITE_PALETTE_SLOT << 12) ;
  finger_sprite.attribute1 =
    finger_sprite.attribute1 | Selection_Dir_Column1_X ;
  finger_sprite.attribute0 = finger_sprite.attribute0 | Selection_Dir_Line1_Y ;
  while (REG_VCOUNT < 160) ;
  memcpy (&sprite_video_memory[0], &finger_sprite, sizeof (struct sprite)) ;

  /* Oh, que c'est encore crad ici. Bon allez, c'est le dernier :)) */
  if (pre_selected_level > -1) {
    level_start = pre_selected_level ;
    goto direct_level_selection ;
  }
  set_screen_mode (MODE_0 | BG0_ENABLE | OBJ_MAP_1D | OBJ_ENABLE) ;

  for (;;) {
    old_cursor_selection = cursor_selection ;
    KEY_POOL ;

    if (KEY_HIT (BUTTON_A)) {
      switch (cursor_selection) {
      case DIR_SELECTION_1_20: level_start = 0 ; break ;
      case DIR_SELECTION_21_40: level_start = 20 ; break ;
      case DIR_SELECTION_41_60: level_start = 40 ; break ;
      case DIR_SELECTION_61_80: level_start = 60 ; break ;
      case DIR_SELECTION_81_100: level_start = 80 ; break ;
      case DIR_SELECTION_101_120: level_start = 100 ; break ;
      case DIR_SELECTION_121_140: level_start = 120 ; break ;
      case DIR_SELECTION_141_160: level_start = 140 ; break ;
      }
    direct_level_selection:
      level_start = handle_level_input (level_start) ;
      if (level_start == -1) {
	pre_selected_level = -1 ;  /* On annule la pr-slection. */
	goto dirty_re_init ;
      }
      else return (level_start) ;
    }

    if ((KEY_HIT (KEYPAD_LEFT)) || (KEY_HIT (KEYPAD_RIGHT))) {
      /* Dplacement vers la gauche ou vers la droite. */
      switch (cursor_selection) {
      case DIR_SELECTION_1_20:
      case DIR_SELECTION_21_40:
      case DIR_SELECTION_41_60:
      case DIR_SELECTION_61_80:
	cursor_selection += 4 ;
	break ;

      case DIR_SELECTION_81_100:
      case DIR_SELECTION_101_120:
      case DIR_SELECTION_121_140:
      case DIR_SELECTION_141_160:
	cursor_selection -= 4 ;
	break ;
      }
    }

    if (KEY_HIT (KEYPAD_UP)) {
      /* Dplacement vers le haut. */
      switch (cursor_selection) {
      case DIR_SELECTION_1_20: cursor_selection = DIR_SELECTION_61_80 ; break ;
      case DIR_SELECTION_21_40: cursor_selection = DIR_SELECTION_1_20 ; break ;
      case DIR_SELECTION_41_60: cursor_selection = DIR_SELECTION_21_40 ;
	break ;
      case DIR_SELECTION_61_80: cursor_selection = DIR_SELECTION_41_60 ;
	break ;

      case DIR_SELECTION_81_100 : cursor_selection = DIR_SELECTION_141_160 ;
	break ;
      case DIR_SELECTION_101_120: cursor_selection = DIR_SELECTION_81_100 ;
	break ;
      case DIR_SELECTION_121_140: cursor_selection = DIR_SELECTION_101_120 ;
	break ;
      case DIR_SELECTION_141_160: cursor_selection = DIR_SELECTION_121_140 ;
	break ;
      }
    }
    
    if (KEY_HIT (KEYPAD_DOWN)) {
      /* Dplacement vers le bas. */
      switch (cursor_selection) {
      case DIR_SELECTION_1_20: cursor_selection = DIR_SELECTION_21_40 ; break ;
      case DIR_SELECTION_21_40: cursor_selection = DIR_SELECTION_41_60 ;
	break ;
      case DIR_SELECTION_41_60: cursor_selection = DIR_SELECTION_61_80 ;
	break ;
      case DIR_SELECTION_61_80: cursor_selection = DIR_SELECTION_1_20 ;
	break ;

      case DIR_SELECTION_81_100 : cursor_selection = DIR_SELECTION_101_120 ;
	break ;
      case DIR_SELECTION_101_120: cursor_selection = DIR_SELECTION_121_140 ;
	break ;
      case DIR_SELECTION_121_140: cursor_selection = DIR_SELECTION_141_160 ;
	break ;
      case DIR_SELECTION_141_160: cursor_selection = DIR_SELECTION_81_100 ;
	break ;
      }
    }

    /* On change la position du curseur seulement s'il a boug. */
    if (cursor_selection != old_cursor_selection) {
      u16 x = 0, y = 0 ;
      switch (cursor_selection) {
      case DIR_SELECTION_1_20:
	x = Selection_Dir_Column1_X ; y = Selection_Dir_Line1_Y ; break ;
      case DIR_SELECTION_21_40:
	x = Selection_Dir_Column1_X ; y = Selection_Dir_Line2_Y ; break ;
      case DIR_SELECTION_41_60:
	x = Selection_Dir_Column1_X ; y = Selection_Dir_Line3_Y ; break ;
      case DIR_SELECTION_61_80:
	x = Selection_Dir_Column1_X ; y = Selection_Dir_Line4_Y ; break ;

      case DIR_SELECTION_81_100:
	x = Selection_Dir_Column2_X ; y = Selection_Dir_Line1_Y ; break ;
      case DIR_SELECTION_101_120:
	x = Selection_Dir_Column2_X ; y = Selection_Dir_Line2_Y ;break ;
      case DIR_SELECTION_121_140:
	x = Selection_Dir_Column2_X ; y = Selection_Dir_Line3_Y ;break ;
      case DIR_SELECTION_141_160:
	x = Selection_Dir_Column2_X ; y = Selection_Dir_Line4_Y ;break ;
      }

      finger_sprite.attribute1 = finger_sprite.attribute1 & 0xFE00 ;
      finger_sprite.attribute1 = finger_sprite.attribute1 | x ;
      finger_sprite.attribute0 = finger_sprite.attribute0 & 0xFF00 ;
      finger_sprite.attribute0 = finger_sprite.attribute0 | y ;  /* y */
      /* Transfert des infos de sprite dans la mmoire vido-sprite. */
      while (REG_VCOUNT < 160) ;
      memcpy (sprite_video_memory, &finger_sprite, sizeof (struct sprite)) ;
    }
  }
}





int setup_directory_selection (int pre_selected_level)
{

  set_screen_mode (0) ;

  /* Chargement des tiles, de la map et de la palette des directories. */
  load_palette_tile_map_256
    (CHAR_BASE_BLOCK_0_ADDR,       /* Tiles dest. */
     SCREEN_BASE_BLOCK_8_ADDR,     /* Map dest. */
     LEV_SEL_PALETTE_SIZE,          /* Taille palette. */
     lev_sel_palette,              /* Palette src. */
     LEV_SEL_NB_TILES,             /* Nb tiles. */
     lev_sel_tiles,                /* Tiles src. */
     LEV_SEL_MAP_WIDTH,            /* Largeur map. */
     (LEV_SEL_MAP_HEIGHT / 2),     /* Hauteur map. */
     lev_sel_map) ;                /* Map src. */

  /* Chargement des de la map de fond des levels. */
  load_palette_tile_map_256
    (CHAR_BASE_BLOCK_0_ADDR,       /* Tiles dest. */
     SCREEN_BASE_BLOCK_9_ADDR,     /* Map dest. */
     0,                            /* Taille palette. */
     NULL,                         /* Palette src. */
     0,                            /* Nb tiles. */
     0,                            /* Tiles src. */
     LEV_SEL_MAP_WIDTH,            /* Largeur map. */
     (LEV_SEL_MAP_HEIGHT / 2),     /* Hauteur map. */
     &lev_sel_map
       [(LEV_SEL_MAP_HEIGHT / 2) * LEV_SEL_MAP_WIDTH]) ;  /* Map src. */

  /* Les sprites sont dj chargs donc le pointeur */
  /* en forme de doigt est dj disponible.         */
  REG_BG0HOFS = 0 ;
  REG_BG0VOFS = 0 ;
  return (handle_directory_input (pre_selected_level)) ;
}
