#include "shapes.h"


/* Rotations customs:
     x' = x cos a - y sin a
     y' = x sin a + y cos a

        | 0 | 90 | 180 | -90
    ------------------------
    cos | 1 | 0  |  -1 |   0
    sin | 0 | 1  |   0 |  -1
*/



static void deltas_rotate_90 (struct deltas *deltas)
{
  unsigned short r_x, r_y ;

  r_x = - deltas->y_delta ;
  r_y = deltas->x_delta ;
  deltas->x_delta = r_x ;
  deltas->y_delta = r_y ;
}





static void deltas_rotate_270 (struct deltas *deltas)
{
  unsigned short r_x, r_y ;

  r_x = deltas->y_delta ;
  r_y = - deltas->x_delta ;
  deltas->x_delta = r_x ;
  deltas->y_delta = r_y ;
}





static void deltas_rotate_180 (struct deltas *deltas)
{
  unsigned short r_x, r_y ;

  r_x = - deltas->x_delta ;
  r_y = - deltas->y_delta ;
  deltas->x_delta = r_x ;
  deltas->y_delta = r_y ;
}


/* Dfinition des 3 fonctions de rotation des pices par stringification. */
#define SHAPE_ROTATE_TEMPLATE(ID_ROT) \
void shape_rotate_##ID_ROT (struct chain_code *shape) \
{ \
  unsigned short i ; \
  \
  for (i = 0; i < shape->length; i++) \
    deltas_rotate_##ID_ROT (&shape->codes[i]) ; \
}



SHAPE_ROTATE_TEMPLATE(90)
SHAPE_ROTATE_TEMPLATE(180)
SHAPE_ROTATE_TEMPLATE(270)



void gen_chain_code_c_source (const char* name, struct chain_code *shape)
{
  int i ;

  printf ("struct deltas_%s_??[] \n  { ", name) ;

  for (i = 0; i < shape->length; i++) {
    printf ("{ %d, %d }, ", shape->codes[i].x_delta, shape->codes[i].y_delta) ;
  }

 printf (" };\n") ;
 printf ("const struct chain_code %s = {\n\
  %d,                         /* length */\n\
  deltas_%s_??    /* codes */\n\
};\n", name, shape->length, name) ;
}



/* Seulement si l'on veut tester sur un terminal style Unix. */
#ifdef CONSOLE_DEBUG
#define LINE_PATTERN "|          |"

void print_shape (struct chain_code *shape)
{
  short x = 4 ;
  short y = 4 ;
  unsigned short i ;
  unsigned char *line ;
  unsigned char *buffer[10] ;


  for (i = 0; i < 10; i++) {
    buffer[i] = (unsigned char*) malloc (1 + sizeof (LINE_PATTERN)) ;
    strcpy (buffer[i], LINE_PATTERN) ;
  }

  for (i = 0; i < shape->length; i++) {
    x = x + shape->codes[i].x_delta ;
    y = y + shape->codes[i].y_delta ;
    line = buffer[y] ;
    line[x] = '.' ;
  }

  for (i = 0; i < 10; i++) {
    printf ("%2i %s\n", i, buffer[i]) ;
    free (buffer[i]) ;
  }
}
#endif
