#include <string.h>
#include <gba.h>
#include <screen_mode.h>
#include "load_gfx_data.h"
#include "key_control.h"
#include "pause_menu_gfx.h"
#include "arrow_sprite_gfx.h"
#include "pause.h"



static struct sprite arrow_sprite = { 0, 0, 0, 0 } ;





static short handle_pause_input ()
{
  u8 cursor_selection, old_cursor_selection ;
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;  

  cursor_selection = PAUSE_SELECTION_CONTINUE ;
  for (;;) {
    old_cursor_selection = cursor_selection ;

    /* Le bouton B fait reprendre par dfaut. */
    KEY_POOL ;
    if (KEY_HIT (BUTTON_B)) return (PAUSE_SELECTION_CONTINUE) ;
    if (KEY_HIT (BUTTON_A)) return (cursor_selection) ;

    if (KEY_HIT (KEYPAD_UP)) {
      /* Dplacement vers le haut. */
      if (cursor_selection == PAUSE_SELECTION_CONTINUE)
	cursor_selection = PAUSE_SELECTION_ABANDON ;
      else cursor_selection-- ;
    }
    
    if (KEY_HIT (KEYPAD_DOWN)) {
      /* Dplacement vers le bas. */
      cursor_selection = (cursor_selection + 1) % 3 ;
    }

    /* On change la position du curseur seulement s'il a boug. */
    if (cursor_selection != old_cursor_selection) {
      u16 y = ARROW_SPRITE_CONTINUE_Y ;
      switch (cursor_selection) {
      case PAUSE_SELECTION_CONTINUE:
	y = ARROW_SPRITE_CONTINUE_Y ; break ;
      case PAUSE_SELECTION_RESTART:
	y = ARROW_SPRITE_RETRY_Y ; break ;
      case PAUSE_SELECTION_ABANDON:
	y = ARROW_SPRITE_ABANDON_Y ; break ;
      }
      arrow_sprite.attribute0 = arrow_sprite.attribute0 & 0xFF00 ;
      arrow_sprite.attribute0 = arrow_sprite.attribute0 | y ;  /* y */
      /* Transfert des infos de sprite dans la mmoire vido-sprite. */
      while (REG_VCOUNT < 160) ;
      memcpy (sprite_video_memory, &arrow_sprite, sizeof (struct sprite)) ;
    }
  }
}





short invoke_pause_menu ()
{
  u16 old_screen_mode ;
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;
  short ret_val ;

  /* Initialisations pour le background 0 (layout de base). */
  set_background2
    (BACKGROUND_PRIORITY_1 |            /* Priorit haute. */
     BACKGROUND_CHAR_BASE_BLOCK_3 |     /* Tiles */
     BACKGROUND_COLOR_256 |             /* 256 couleurs. */
     BACKGROUND_SCREEN_BASE_BLOCK_10 |  /* Map. */
     TEXT_BACKGROUND_256x256) ;         /* 256 x 256 pixels. */

  /* Chargement des tiles et de la map de fond d'cran de jeu. */
  load_palette_tile_map_256
    (CHAR_BASE_BLOCK_3_ADDR,           /* Tiles dest. */
     SCREEN_BASE_BLOCK_10_ADDR,        /* Map dest. */
     PAUSE_MENU_PALETTE_SIZE,          /* Taille palette. */
     pause_menu_palette,               /* Palette src. */
     PAUSE_MENU_NB_TILES,              /* Nb tiles. */
     pause_menu_tiles,                 /* Tiles src. */
     PAUSE_MENU_MAP_WIDTH,             /* Largeur map. */
     PAUSE_MENU_MAP_HEIGHT,            /* Hauteur map. */
     pause_menu_map) ;                 /* Map src. */

  REG_BG2HOFS = 0 ;
  REG_BG2VOFS = 0 ;

  load_sprites_16 (ARROW_SPRITE_LOADING_OFFSET,  /* Offset. */
		   ARROW_SPRITE_NB_TILES,        /* Nb tiles. */
		   arrow_sprite_tiles,           /* Tiles src. */
		   ARROW_SPRITE_PALETTE_SLOT,    /* Slot palette. */
		   ARROW_SPRITE_PALETTE_SIZE,    /* Taille palette. */
		   arrow_sprite_palette) ;       /* Palette src. */

  /* Initialisation du sprite du curseur en forme de flche. */
  arrow_sprite.attribute0 = SPRITE_16_COLOR | SPRITE_SQUARE ;
  arrow_sprite.attribute1 = SPRITE_SIZE_16 ;
  arrow_sprite.attribute2 =
    ARROW_SPRITE_TILE_INDEX | (ARROW_SPRITE_PALETTE_SLOT << 12) ;
  arrow_sprite.attribute1 = arrow_sprite.attribute1 | ARROW_SPRITE_X ;
  arrow_sprite.attribute0 = arrow_sprite.attribute0 | ARROW_SPRITE_CONTINUE_Y ;
  while (REG_VCOUNT < 160) ;
  memcpy (&sprite_video_memory[0], &arrow_sprite, sizeof (struct sprite)) ;

  old_screen_mode = REG_DISPCNT ;
  /* On rajoute la visibilit du background 2. */
  set_screen_mode
    (REG_DISPCNT | MODE_0 | BG2_ENABLE | OBJ_MAP_1D | OBJ_ENABLE) ;

  ret_val = handle_pause_input () ;
  set_screen_mode (old_screen_mode) ;
  return (ret_val) ;
}
