#ifndef __SHAPES__H_
#define __SHAPES__H_


/* Formes des pices. */
enum shape {
  Shape_None = 0x00,
  Shape_Square_1x1 = 0x01,
  /* U/D/L/R
      _
     |_|
  */
  Shape_Square_4x4 = 0x02,
  /* U/D/L/R
      __
     |  |
     |__|
  */
  Shape_Long = 0x03,
  /*  U/D  L/R
       _
      | |  ____
      | | |____|
      | |
      |_|
  */
  Shape_Junction = 0x04,
  /*  R       L     U       D
       _       _     _     _____
      | |_   _| |  _| |_  |_   _|
      |  _| |_  | |_____|   |_|
      |_|     |_|
  */
  Shape_Stair_1 = 0x05,
  /*   R/L     U/D
       ___        _
      |_  |_    _| |
        |___|  |  _|
               |_|
  */
  Shape_Stair_2 = 0x06,
   /*    ___    _
       _|  _|  | |_
      |___|    |_  |
                 |_|
   */
  Shape_Corner = 0x07
  /*  _
     | |_
     |___|
   */
};
#define SHAPES_NUMBER 7

#define Default_Shape_Square_1x1_Hot_Spot 0
#define Default_Shape_Square_4x4_Hot_Spot 0
#define Default_Shape_Long_Hot_Spot 1
#define Default_Shape_Junction_Hot_Spot 1
#define Default_Shape_Stair_1_Hot_Spot 1
#define Default_Shape_Stair_2_Hot_Spot 1
#define Default_Shape_Corner_Hot_Spot 1

/* Orientation de pice. */
enum orientation {
  Orient_Up = 0x0,
  Orient_Right = 0x1,
  Orient_Down = 0x2,
  Orient_Left = 0x3
};



/* Structure des pices disposes sur l'aire de jeu. */
/* Structure de piece:
   Bits:
   15 - 8 : Numro de la part de la pice.
   7 - 4 : Orientation de la pice.
   3 - 0 : Forme de la pice.
 */
typedef unsigned short Piece ;



#define GET_SHAPE(piece) ((enum shape) ((piece) & 0x000F))

#define SET_ORIENT(piece,orient) \
  (piece) = (((piece) & 0xFF0F) | (((unsigned short) (orient)) << 4))
#define GET_ORIENT(piece) ((enum orientation) (((piece) & 0x00F0) >> 4))

#define GET_PART(piece) ((short) (((piece) & 0xFF00)) >> 8)
#define SET_PIECE(piece,partnum,orient,shape) \
  (piece) = ((((unsigned short) partnum) << 8) & 0xFF00) | \
            ((((unsigned short) orient) << 4) & 0x00F0) | \
            (((unsigned short) shape) & 0x000F)
#define UNSET_PIECE(piece) (piece) = ((Piece) 0x0000)




/* Chain-codes reprsentant la forme des pices. */
struct deltas {
  short x_delta ;
  short y_delta ;
};

struct chain_code {
  unsigned short length ;
  struct deltas *codes ;
};


struct chain_code* copy_shape (struct chain_code *) ;
void free_shape (struct chain_code *) ;

#endif
