#include <string.h>
#include <gba.h>
#include <screen_mode.h>
#include <interrupt.h>
#include "load_gfx_data.h"
#include "it_handlers.h"
#include "intro_gfx.h"
#include "key_control.h"
#include "intro_sprite_gfx.h"



#define BLEND_DELAY 16

void show_at_toplevel (u16 palette_size,
		       const u16 *palette,
		       u16 nb_tiles,
		       const u8 *tiles,
		       u16 map_width,
		       u16 map_height,
		       const u16 *map)
{
  struct sprite *sprite_video_memory = SPRITES_INFO_MEMORY_ADR ;
  struct sprite intro_sprite = { 0, 0, 0, 0 } ;
  unsigned long previous_elapsed_seconds = 0 ;
  u8 pressed = 0 ;
  u16 i ;

  set_screen_mode (0) ;

  set_background0
    (BACKGROUND_PRIORITY_3 |            /* Priorit la plus faible. */
     BACKGROUND_CHAR_BASE_BLOCK_0 |     /* Tiles */
     BACKGROUND_COLOR_256 |             /* 256 couleurs. */
     BACKGROUND_SCREEN_BASE_BLOCK_8 |   /* Map. */
     TEXT_BACKGROUND_256x256) ;         /* 256 x 256 pixels. */

  load_palette_tile_map_256
    (CHAR_BASE_BLOCK_0_ADDR,
     SCREEN_BASE_BLOCK_8_ADDR,
     palette_size,
     palette,
     nb_tiles,
     tiles,
     map_width,
     map_height,
     map) ;

  REG_BG0HOFS = 0 ;
  REG_BG0VOFS = 0 ;

  load_sprites_16 (INTRO_SPRITE_LOADING_OFFSET,  /* Offset. */
		   INTRO_SPRITE_NB_TILES,        /* Nb tiles. */
		   intro_sprite_tiles,           /* Tiles src. */
		   INTRO_SPRITE_PALETTE_SLOT,    /* Slot palette. */
		   INTRO_SPRITE_PALETTE_SIZE,    /* Taille palette. */
		   intro_sprite_palette) ;       /* Palette src. */
  /* Initialisation du sprite du curseur. */
  intro_sprite.attribute0 = SPRITE_16_COLOR | SPRITE_SQUARE ;
  intro_sprite.attribute1 = SPRITE_SIZE_32 ;
  intro_sprite.attribute2 =
    INTRO_SPRITE_RELEASED_TILE_INDEX | (INTRO_SPRITE_PALETTE_SLOT << 12) ;
  intro_sprite.attribute1 = intro_sprite.attribute1 | INTRO_SPRITE_X ;
  intro_sprite.attribute0 = intro_sprite.attribute0 | INTRO_SPRITE_Y ;
  while (REG_VCOUNT < 160) ;
  memcpy (&sprite_video_memory[0], &intro_sprite, sizeof (struct sprite)) ;

  REG_COLEY = 0 ;
  set_blending_mode (FADE_MODE_2 | FADE_SOURCE_BG1) ;
  set_screen_mode (MODE_0 | BG0_ENABLE | OBJ_MAP_1D | OBJ_ENABLE) ;
  for (i = 0; i < (16 * BLEND_DELAY); i++) {
    while (REG_VCOUNT != 160) ;
    REG_COLEY = (u16) (15 - (i / BLEND_DELAY)) ;
  }
  set_blending_mode (0) ;

  /* Mainloop d'attente. */
  KEY_POOL ;
  start_timer2 () ;
  for (;;) {
    if (previous_elapsed_seconds < elapsed_seconds) {
      previous_elapsed_seconds = elapsed_seconds ;
      set_screen_mode (REG_DISPCNT & ~(OBJ_MAP_1D | OBJ_ENABLE)) ;

      if (pressed == 1) {
	pressed = 0 ;
	intro_sprite.attribute2 =
	  INTRO_SPRITE_RELEASED_TILE_INDEX |
	  (INTRO_SPRITE_PALETTE_SLOT << 12) ;
      }
      else {
	pressed = 1 ;
	intro_sprite.attribute2 =
	  INTRO_SPRITE_PRESSED_TILE_INDEX | (INTRO_SPRITE_PALETTE_SLOT << 12) ;
      }
      while (REG_VCOUNT < 160) ;
      memcpy (&sprite_video_memory[0], &intro_sprite, sizeof (struct sprite)) ;
      set_screen_mode (REG_DISPCNT | OBJ_MAP_1D | OBJ_ENABLE) ;
    }
    KEY_POOL ;
    if ((KEY_HIT (BUTTON_A)) || (KEY_HIT(BUTTON_SELECT)) ||
	(KEY_HIT (BUTTON_START))) break ;
  }

  stop_timer2 () ;
  intro_sprite.attribute1 = intro_sprite.attribute1 & 0xFE00 ;
  intro_sprite.attribute0 = intro_sprite.attribute0 & 0xFF00 ;
  intro_sprite.attribute1 = intro_sprite.attribute1 | 240 ;
  intro_sprite.attribute0 = intro_sprite.attribute0 | 160 ;
  while (REG_VCOUNT < 160) ;
  memcpy (&sprite_video_memory[0], &intro_sprite, sizeof (struct sprite)) ;
}
